using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;

namespace ParseAnsi
{
	/// <summary>
	/// Summary description for AnsiPump.
	/// </summary>
	public class AnsiPump
	{
		delegate void PumpDelegate();
		delegate void ParseControlSequenceDelegate(string command);

		Process process;
		TextReader input;
		TextWriter output;
		IAsyncResult asyncResult;
		IntPtr consoleHandle;
		ParseControlSequenceDelegate parseControlSequenceHandler;

		public AnsiPump(Process Process, TextReader Input, TextWriter Output, int StdHandle)
		{
			process = Process;
			input = Input;
			output = Output;
			consoleHandle = Win32API.GetStdHandle(StdHandle);
			Win32API.GetConsoleScreenBufferInfo(consoleHandle, ref defaultinfo);
			parseControlSequenceHandler = new ParseControlSequenceDelegate(parseControlSequence);
			PumpDelegate pump = new PumpDelegate(Pump);
			asyncResult = pump.BeginInvoke(null, null);
		}

		public void WaitForExit()
		{
			asyncResult.AsyncWaitHandle.WaitOne();
		}

		// ANSI Parsing coded from http://www.inwap.com/pdp10/ansicode.txt
		private char[] ANSI_Intermediate = " !\"#$%&'()*+,-./".ToCharArray();
		private char[] ANSI_Parameters = "0123456789:;<=>?".ToCharArray();
		private char[] ANSI_Uppercase = "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_".ToCharArray();
		private char[] ANSI_Lowercase = "`abcdefghijlkmnopqrstuvwxyz{|}~".ToCharArray();
		private char[] ANSI_Alphabetic = "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijlkmnopqrstuvwxyz{|}~".ToCharArray();

		const int BUFFERSIZE = 4096; // Default size of stream buffer
		char[] buffer = new char[BUFFERSIZE];
		int bufferstart = 0;
		char[] specials = new char[] { (char)27, (char)12 };
		Win32API.CONSOLE_SCREEN_BUFFER_INFO defaultinfo = new Win32API.CONSOLE_SCREEN_BUFFER_INFO();

		public void Pump()
		{
			while (true)
			{
				try
				{
					if (!doReadAndParse()) break;
				}
				catch (Exception e)
				{
					output.WriteLine(e.ToString());
				}
			}
		}

		private bool doReadAndParse()
		{
			if (BUFFERSIZE - bufferstart == 0)
				bufferstart = 0;
			int read = input.Read(buffer, bufferstart, BUFFERSIZE - bufferstart);
			if (read == 0) return false;

			read += bufferstart;
			string toparse = new string(buffer, 0, read);

			int parsestart = 0;
			while (parsestart < read)
			{
				int firstat = toparse.IndexOfAny(specials, parsestart, read - parsestart);

				if (firstat > -1)
				{
					switch (buffer[firstat])
					{
						case (char)27:

							output.Write(buffer, parsestart, firstat - parsestart);
							output.Flush();

							if (firstat + 1 >= read)
							{
								queueForNextRead(firstat, read - firstat);
								return true;
							}

							char nextchar = buffer[firstat + 1];

							switch (nextchar)
							{
								case '[': // CSI - Control Sequence - Ends with alphabetic character
									int alphaat = toparse.IndexOfAny(ANSI_Alphabetic, firstat + 2, read - firstat - 2);
									if (alphaat < 0)
									{
										queueForNextRead(firstat, read - firstat);
										return true;
									}
									string sequence = toparse.Substring(firstat + 2, alphaat - firstat - 1);
									parseControlSequenceHandler(sequence);
									parsestart = alphaat + 1;
									break;
								default:
									parsestart = firstat + 1;
									output.Write("Unknown ANSI escape sequence");
									break;
							}
							break;
						case (char)12:
							parseControlSequenceHandler("2J");
							parsestart = firstat + 1;
							break;
					}
				} 
				else 
				{
					output.Write(buffer, parsestart, read - parsestart);
					output.Flush();
					break;
				}
			}

			bufferstart = 0;
			return true;
		}

		private void queueForNextRead(int endofparse, int length)
		{
			Array.Copy(buffer, endofparse, buffer, 0, length);
			bufferstart = length;
		}

		private void parseControlSequence(string command)
		{
			lock (process)
			{
				int intpos = command.IndexOfAny(ANSI_Intermediate);
				if (intpos == -1) intpos = command.Length;
				string paramstr = command.Substring(0, intpos - 1);
				string[] paramstrs = paramstr.Split(';');
				short distance = 1;
				Win32API.CONSOLE_SCREEN_BUFFER_INFO info = new Win32API.CONSOLE_SCREEN_BUFFER_INFO();
				Win32API.GetConsoleScreenBufferInfo(consoleHandle, ref info);
				switch (command[command.Length - 1])
				{
					case 'A':
						if (paramstrs[0].Length > 0) distance = Convert.ToInt16(paramstrs[0]);
						info.dwCursorPosition.y -= distance;
						Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						break;
					case 'B':
						if (paramstrs[0].Length > 0) distance = Convert.ToInt16(paramstrs[0]);
						info.dwCursorPosition.y += distance;
						Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						break;
					case 'C':
						if (paramstrs[0].Length > 0) distance = Convert.ToInt16(paramstrs[0]);
						info.dwCursorPosition.x += distance;
						Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						break;
					case 'D':
						if (paramstrs[0].Length > 0) distance = Convert.ToInt16(paramstrs[0]);
						info.dwCursorPosition.x -= distance;
						Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						break;
					case 'H':
						short row = 1;
						if (paramstrs[0].Length > 0) row = Convert.ToInt16(paramstrs[0]);
						if (paramstrs.Length > 1) distance = Convert.ToInt16(paramstrs[1]);
						info.dwCursorPosition.y = (short)(row - 1);
						info.dwCursorPosition.x = (short)(distance - 1);
						Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						break;
					case 'J':
						distance = 0;
						if (paramstrs[0].Length > 0) distance = Convert.ToInt16(paramstrs[0]);
						if (distance == 2) 
						{
							int hWrittenChars = 0;
							info.dwCursorPosition.x = info.dwCursorPosition.y = 0;
							Win32API.FillConsoleOutputCharacter(consoleHandle, 32, info.dwSize.x * info.dwSize.y, info.dwCursorPosition, ref hWrittenChars);
							Win32API.FillConsoleOutputAttribute(consoleHandle, info.wAttributes, info.dwSize.x * info.dwSize.y, info.dwCursorPosition, ref hWrittenChars);
							Win32API.SetConsoleCursorPosition(consoleHandle, info.dwCursorPosition);
						}
						else
						{
							output.Write("Unknown clear sequence: " + command);
						}
						break;
					case 'm':
						Win32API.CharacterAttributes style = info.wAttributes;
						foreach (string param in paramstrs)
						{
							int val = Convert.ToInt32(param);
							if (val == 0)
							{
								style = defaultinfo.wAttributes;
							} 
							else if (val == 1)
							{
								style |= Win32API.CharacterAttributes.FOREGROUND_INTENSITY;
							} 
							else if (val == 4)
							{
								style |= Win32API.CharacterAttributes.COMMON_LVB_UNDERSCORE;
							} 
							else if (val == 7)
							{
								style |= Win32API.CharacterAttributes.COMMON_LVB_REVERSE_VIDEO;
							} 
							else if ((val >= 30) && (val <= 37))
							{
								val -= 30;
								val = ((val & 0x1) << 2) | (val & 0x2) | ((val & 0x4) >> 2);
								style &= (Win32API.CharacterAttributes)0xFFF8;
								style |= (Win32API.CharacterAttributes)val;
							} 
							else if ((val >= 40) && (val <= 47))
							{
								val -= 40;
								val = ((val & 0x1) << 2) | (val & 0x2) | ((val & 0x4) >> 2);
								val <<= 4;
								style &= (Win32API.CharacterAttributes)0xFF8F;
								style |= (Win32API.CharacterAttributes)val;
							}
							else
							{
								output.Write("Unknown color sequence: " + command);
							}
						}
						Win32API.SetConsoleTextAttribute(consoleHandle, style);
						break;
					default:
						output.Write("Unknown control sequence: " + command);
						break;
				}
			}
		}
	}
}
